/* ====================================================
 * CUTEst interface for cg_descent       May 1, 2018
 *
 * W. Hager
 *
 * (Based on CUTEr gencma.c of D. Orban, Feb 3, 2003)
 * (CUTEst evolution, Nick Gould, Apr 2, 2014)
 * ====================================================
 */

/* macros */
#define CG_DESCENTMA

#define MAXLINE 256

#ifdef __cplusplus
extern "C" {   /* To prevent C++ compilers from mangling symbols */
#endif

#include "cutest.h"
#include "cg_descent.h"

#ifdef Isg95
#define MAINENTRY MAIN_
#else
#define MAINENTRY main
#endif

void cg_value
(
    CGFLOAT *f,
    CGFLOAT *x,
    CGINT    n
) ;

void cg_grad
(
    CGFLOAT  *g,
    CGFLOAT  *x,
    CGINT     n
) ;

void cg_valgrad
(
    CGFLOAT  *f,
    CGFLOAT  *g,
    CGFLOAT  *x,
    CGINT     n
) ;

void hessian_at_x
(
    SOPT_matrix *H, /* see sopt.h for the structure SOPT_matrix */
    CGFLOAT     *x, /* the point where the Hessian is evaluated */
    CGINT        n  /* dimension of x */
) ;

void cute_cg_error
(
    int status  /* error status */
) ;

/* global variables */
    integer cutest_nvar, cutest_nnzh, *cutest_H_row, *cutest_H_col ;
    logical cute_true = TRUE_ ;
    CGFLOAT *cutest_H_val ;
    
/* main program */
int MAINENTRY( void )
{
    double walltime ; /* wall time  */
    char *fname = "OUTSDIF.d"; /* CUTEst data file */
    integer funit = 42;        /* FORTRAN unit number for OUTSDIF.d */
    integer io_buffer = 11;    /* FORTRAN unit for internal i/o */
    integer iout = 6;          /* FORTRAN unit number for error output */
    integer ierr;              /* Exit flag from OPEN and CLOSE */
    integer status;            /* Exit flag from CUTEst tools */
    char fullpathname [1024] ;
    char *cutest_dir ;

    integer     anz ;
    CGFLOAT     *x, *lo, *hi, *zeros ;
    char        qpprob [MAXLINE+1], uprob [MAXLINE+1], *uname, *qpname, *pname ;

    doublereal  calls[7], cpu[2] ;
    integer     ExitCode;
    int         QP, status_cg, U ;
    CGFLOAT     s, t ;
    CGINT       i, ncol ;

    FILE *QPfile, *Ufile ;

    /* for an LP or QP, fadjust stores constant term in objective */
    CGFLOAT fadjust = CGZERO ;

    CGFLOAT *c = NULL ; /* linear cost vector */

    /* Open problem description file OUTSDIF.d */
    ierr = 0;
    FORTRAN_open( &funit, fname, &ierr ) ;
    if( ierr )
    {
        printf("Error opening file OUTSDIF.d.\nAborting.\n") ;
        exit(1) ;
    }

    /* --- Setup for unconstrained problem ---------- */
    printf ("Initializing CGdata structure.\n") ;
    CGdata *cgdata = cg_setup () ;
    if ( cgdata == NULL )
    {
            cg_error (-1, __FILE__, __LINE__, "cg_setup failed") ;
    }
    else
    {
        printf ("Successfully initialized CGdata structure.\n") ;
    }
    CGparm *Parm = cgdata->Parm ;
    CGstat *Stat = cgdata->Stat ;

    SOPT_matrix *H = cgdata->H ;

    /* Determine problem size */
    CUTEST_udimen( &status, &funit, &cutest_nvar) ;
    if (status)
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status);
        exit(status);
    }
    ncol = cutest_nvar ; /* CGint */

    /* Reserve memory for variables, bounds, and multipliers */
    /* and call appropriate initialization routine for CUTEst */
    x =  (CGFLOAT *) malloc (ncol*sizeof (CGFLOAT)) ;
    lo = (CGFLOAT *) malloc (ncol*sizeof (CGFLOAT)) ;
    hi = (CGFLOAT *) malloc (ncol*sizeof (CGFLOAT)) ;
    CUTEST_usetup ( &status, &funit, &iout, &io_buffer, &ncol, x, lo, hi ) ;

    if (status)
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status);
        exit(status);
    }

    /* Get problem name */
    pname = (char *) malloc ((FSTRING_LEN+1)*sizeof (char)) ;
    CUTEST_probname( &status, pname ) ;
    if (status)
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status) ;
        exit(status) ;
    }

    /* Make sure to null-terminate problem name */
    pname[FSTRING_LEN] = '\0';
    i = FSTRING_LEN - 1;
    while(i-- > 0 && pname[i] == ' ')
    {
        pname[i] = '\0';
    }

    /* Print problem name */
    printf ("\n Problem: %s (n = %ld)\n", pname, (LONG) ncol ) ;

    /* read the full path to the pasa/CUTEst directory and store it in the
       variable cutest_dir */
#include ".cutest_location"

    /* See if the problem is in the list of unconstrained problem.
       If not, then terminate with an error message. */
    strcpy (fullpathname, cutest_dir) ;
    Ufile = fopen (strcat (fullpathname, "/classU"), "r") ;
    U = FALSE ;
    while (fgets (uprob, MAXLINE, Ufile) != (char *) NULL)
    {
        for (uname = uprob; *uname; uname++)
        {
            if (isspace (*uname)) *uname = '\0' ;
        }
        uname = uprob ;
        if ( strcmp (uname, pname) == 0 )
        {
            U = TRUE ;
            break ;
        }
    }

    if ( U == FALSE )
    {
        printf ("Problem %s was not found in the list of "
                "unconstrained problems in the file %s\n",
                pname, strcat (fullpathname, "/classU")) ;
        printf ("Either the problem is constrained, or the file should be "
                "updated to include this new unconstrained problem.\n") ;
        cg_error (-1, __FILE__, __LINE__, "STOP") ;
    }

    /* see if the problem has a quadratic objective by comparing the name
       pname of the test problem to the names qpprob of the unconstrained
       QPs contained in the file classQPU */
    strcpy (fullpathname, cutest_dir) ;
    QPfile = fopen (strcat (fullpathname, "/classUQP"), "r") ;
    QP = FALSE ;
    while (fgets (qpprob, MAXLINE, QPfile) != (char *) NULL)
    {
        for (qpname = qpprob; *qpname; qpname++)
        {
            if (isspace (*qpname)) *qpname = '\0' ;
        }
        qpname = qpprob ;
        if ( strcmp (qpname, pname) == 0 )
        {
            QP = TRUE ;
            break ;
        }
    }

    /* initial cgdata structure setup */
    cgdata->x = x ;
    cgdata->n = ncol ;

    /* If the problem has a quadratic objective, then extract the
       Hessian matrix and the linear term in the objective. */
    if ( QP )
    {
        printf ("the problem has a quadratic objective\n") ;
        /* nnzh = the number of nonzeros required to store the sparse Hessian 
           matrix in coordinate format */
        if (status)
        {
            printf("** CUTEst error, status = %ld, aborting\n", (LONG) status) ;
            exit(status);
        }

        /* Reserve memory for H_val, H_row, H_col */
        CUTEST_udimsh ( &status, &cutest_nnzh );

        /* malloc H_val, cutest_H_row, cutest_H_col */
        cutest_H_val = (CGFLOAT *) malloc (cutest_nnzh*sizeof (CGFLOAT)) ;
        cutest_H_row = (integer *) malloc (cutest_nnzh*sizeof (integer)) ;
        cutest_H_col = (integer *) malloc (cutest_nnzh*sizeof (integer)) ;

        /* Store zero in an array named zeros, it is used below. */
        zeros = (CGFLOAT *) malloc (ncol*sizeof (CGFLOAT)) ;
        cg_initx (zeros, CGZERO, ncol) ;

        /* Determine the nonzero values in the Hessian */
        CUTEST_ush (&status, &cutest_nvar, zeros, &anz, &cutest_nnzh,
                         cutest_H_val, cutest_H_row, cutest_H_col) ;

        if (status)
        {
            printf("** CUTEst error, status = %ld, aborting\n", (LONG) status) ;
            exit(status) ;
        }
        if ( cutest_nnzh != anz )
        {
            printf ("nnzh (%ld) != anz (%ld)\n",
                                              (LONG) cutest_nnzh, (LONG) anz) ;
            cg_error (-1, __FILE__, __LINE__, "STOP") ;
        }

        H->rows = cutest_H_row ;
        H->cols = cutest_H_col ;
        H->vals = cutest_H_val ;
        H->nnz  = anz ;
        H->sym  = TRUE ;

        status = sopt_convert_to_sparse (H, ncol, ncol, TRUE, TRUE, LCUTE) ;
        if ( status > 0 )
        {
            cute_cg_error (status) ;
            cg_error (-1, __FILE__, __LINE__, "execution halted") ;
        }

        /* Determine c in objective function */
        c = (CGFLOAT *) malloc (ncol*sizeof (CGFLOAT)) ;
        CUTEST_uofg (&status, &ncol, zeros, &fadjust, c, &cute_true);
        /* store the linear cost vector and the sparse Hessian matrix in
           the cgdata structure */
        cgdata->c = c ;
    }
    else /* the problem is not quadratic */
    {
        cgdata->value = cg_value ;
        cgdata->grad = cg_grad ;
        cgdata->valgrad = cg_valgrad ;
    }

    /* set new parameter values for CG here, otherwise default values are used.
       For example: */

    /* Parm->grad_tol = 1.e-8 ; */ /* default is 1.e-6 */
    /* Parm->deriv_mode = 1 ; */   /* default is 12 */ ;

    /* if cg_descent is used in the Hessian-based mode, then store the routine
       to evaluate the Hessian */
    if ( Parm->deriv_mode > 1 )
    {
        cgdata->hessian = hessian_at_x ;
    }

    /* Call the optimizer */

    /* time run using wall clock, call several times to exclude startup cost */

    walltime = cg_timer () ;
    walltime = cg_timer () ;

    status_cg = cg_descent (cgdata) ;

    walltime = cg_timer () - walltime ;

    ExitCode = 0;

    /* Get CUTEst statistics */
    /* CUTEST_ureport( &status, calls, cpu) ;*/
    if (status)
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status) ;
        exit(status) ;
    }

    /* print unformatted cg_descent statistics */
    /* printf ("!!%10s %6ld %7ld %7ld %7ld %5i %16.7e %16.7e %11.6f\n",
             pname, (LONG) ncol, (LONG) Stat->iter, (LONG) Stat->nfunc,
             (LONG) Stat->ngrad, status_cg, Stat->err, Stat->f+fadjust,
             walltime) ;*/

    printf("\n\n *********************** CG statistics **************"
           "**********\n\n") ;
    printf("Code used                              : cg_descent\n") ;
    printf("Problem                                : %-s\n", pname) ;
    printf("number of variables                    = %-10ld\n\n",
          (LONG) ncol) ;
    printf("number of cg iterations                = %-10ld\n",
          (LONG) Stat->iter) ;
    printf("number of cg function evaluations      = %-10ld\n",
          (LONG) Stat->nfunc) ;
    printf("number of cg gradient evaluations      = %-10ld\n\n",
          (LONG) Stat->ngrad) ;
    printf("Number of times cgexpand called         = %-10ld\n",
          (LONG) Stat->nexpand) ;
    printf("Number of forward expansions            = %-10ld\n",
          (LONG) Stat->nforward) ;
    printf("Number of backward moves                = %-10ld\n\n",
          (LONG) Stat->nback) ;
    printf("Newton steps computed by CG iteration  = %-10ld\n",
          (LONG) Stat->nCG) ;
    printf("Newton steps based on symmetric solver = %-10ld\n",
          (LONG) Stat->nSS) ;
    printf("Number of PRP iterations in Newton CG  = %-10ld\n\n",
          (LONG) Stat->PRP) ;

    printf("|| g ||                                = %-16.7e\n", Stat->err) ;
    printf("Final f                                = %-16.7e\n",
                                                   Stat->f+fadjust) ;
    cg_value (&t, x, cutest_nvar) ;
    printf("Function value at final x              = %-16.7e\n", t) ;
    printf("Solve time                             = %-11.6f seconds\n",
                                                   walltime) ;
    printf("\n ***********************************************************"
           "*******\n\n") ;

    /* Print lines to clearly separate problems */

    printf (" ====================================================\n\n" ) ;

    ierr = 0;
    FORTRAN_close( &funit, &ierr ) ;
    if ( ierr )
    {
        printf( "Error closing %s on unit %ld.\n", fname, (LONG) funit ) ;
        printf( "Trying not to abort.\n" ) ;
    }

    /* Free workspace */
    free (pname) ;
    free (x) ;
    free (lo) ;
    free (hi) ;
    if ( QP )
    {
        free (zeros) ;
        free (H->x) ;
        free (H->i) ;
        free (H->p) ;
        free (c) ;
    }

    cg_terminate (&cgdata) ;

    /* end cutest */
    CUTEST_uterminate( &status ) ;

    /* close files */
    fclose(QPfile) ;
    fclose(Ufile) ;

    return 0;
}

#ifdef __cplusplus
}    /* Closing brace for  extern "C"  block */
#endif
void cg_value
(
    CGFLOAT *f,
    CGFLOAT *x,
    CGINT N
)
{
    integer status, n ;

    n = N ;
    CUTEST_ufn( &status, &n, x, f) ;
    if ((status == 1) || (status == 2))
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status) ;
        exit(status) ;
    }
/*printf ("function value: %e\n", *f) ;*/
    return ;
}

void cg_grad
(
    CGFLOAT *g,
    CGFLOAT *x,
    CGINT N
)
{
    integer n, status;
    n = N ;
    CUTEST_ugr ( &status, &n, x, g) ;
    if ((status == 1) || (status == 2))
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status);
        exit(status);
    }
}

void cg_valgrad
(
    CGFLOAT *f,
    CGFLOAT *g,
    CGFLOAT *x,
    CGINT N
)
{
    integer n, status;
    n = N ;
    CUTEST_uofg( &status, &n, x, f, g, &cute_true) ;
    if ((status == 1) || (status == 2))
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status);
        exit(status);
    }
    return ;
}

/* routine to evaluate the Hessian at a given point x */
void hessian_at_x
(
    SOPT_matrix *H, /* see sopt.h for the structure SOPT_matrix */
    CGFLOAT     *x, /* the point where the Hessian is evaluated */
    CGINT        n  /* dimension of x */
)
{
    integer anz, status ;

    CUTEST_ush (&status, &cutest_nvar, x, &anz, &cutest_nnzh,
                 cutest_H_val, cutest_H_row, cutest_H_col) ;

    if (status)
    {
        printf("** CUTEst error, status = %ld, aborting\n", (LONG) status);
        exit(status);
    }

    H->rows = cutest_H_row ;
    H->cols = cutest_H_col ;
    H->vals = cutest_H_val ;
    H->nnz  = cutest_nnzh ;
    H->sym  = TRUE ;
    return ;
}

void cute_cg_error
(
    int status  /* error status */
)
{
    if ( status == CUTE_OUT_OF_MEMORY )
    {
        printf ("In CGDESCENT/CUTEst/cg_descent_main, ran out of memory.\n") ;
    }
    else if ( status == CUTE_ERROR_IN_INPUT_MATRIX )
    {
        printf ("In CGDESCENT/CUTEst/cg_descent_main, the Hessian of a\n"
                "quadratic was provided.  An error was detected in the input\n"
                "as described above.\n") ;
    }
    else if ( status == CUTE_NROW_OR_NCOL_NOT_GIVEN_FOR_DENSE )
    {
        printf ("In CGDESCENT/CUTEst/cg_descent_main, the coefficient matrix\n"
                "was input as a dense matrix, however, either the number of\n"
                "rows or columns were not provided in the input structure.\n") ;
    }
}
