    /* Test problem: Hock and Schittkowski problem 112 as implemented in CUTEST
                     problem HS112.SIF

         min f(x) s.t. 1.e-6 <= x, A*x = [2 1 1 0]'

       where

         f(x) = sum_{i = 1 to 10} x_i(c_i + log (x_i) - log (x_{11}))

              | 1  2  2  0  0  1  0  0  0  1  0|
          A = | 0  0  0  1  2  1  1  0  0  0  0|
              | 0  0  1  0  0  0  1  1  2  1  0|
              | 1  1  1  1  1  1  1  1  1  1 -1|

          c = [-6.089 -17.164 -34.054  -5.914 -24.721 ...
              -14.986 -24.100 -10.708 -26.662 -22.179 ]

       Starting guess: x_i = 0.1, 1 <= i <= 10, x_{11} = 1

       In CUTEST problem HS112.SIF, the right side [2 1 0 0] of problem 112 is
       replaced by [2 1 1 0]. With this replacement, the solution given by
       Hock and Schittkowski for problem 112 is obtained. */

#include "pasa.h"

/* prototypes */
void value
(
    PASAFLOAT *val,
    PASAFLOAT   *x,
    PASAINT      n
) ;

void grad
(
    PASAFLOAT *g,
    PASAFLOAT *x,
    PASAINT    n
) ;

void valgrad
(
    PASAFLOAT *val,
    PASAFLOAT   *g,
    PASAFLOAT   *x,
    PASAINT      n
) ;

void hess
(
    SOPT_matrix *H,
    SOPTFLOAT   *x,
    SOPTINT      n
) ;

/* global variables */
PASAINT   *rows = NULL ;
PASAINT   *cols = NULL ;
PASAFLOAT *vals = NULL ;
PASAFLOAT c [] = {-6.089, -17.164, -34.054,  -5.914, -24.721,
                 -14.986, -24.100, -10.708, -26.662, -22.179 } ;

/* main function */
int main (void /*int argc, char **argv*/)
{
   /*  Initialize the pasadata structure for storing the problem description.
       pasa_setup returns a pointer to a data structure where
       all parameter values are set to default values. */
    printf ("Initializing PASAdata structure.\n") ;
    PASAdata *pasadata = pasa_setup () ; 
    if ( pasadata == NULL )
    {
        printf ("pasa_setup ran out of memory\n") ;
    }
    else
    {
        printf ("Successfully initialized PASAdata structure.\n") ;
    }

    /* Need to specify the number of columns and rows in A since
       it will be input as a dense matrix by rows and pasa needs to know
       how to extract the rows of the matrix from the array. */
                   pasadata->nrow = 4 ;
    PASAINT ncol = pasadata->ncol = 11 ;

    /* Currently, there are 3 different ways to give the elements of the
       constraint matrix A:

       (1) sparse matrix format: 3 arrays
             Ai - row indices of the nonzeros elements. Indices in each
                  column should be in increasing order.
             Ax - the nonzeros numerical entries corresponding to entries of Ai
             Ap - the array of column pointers (size ncol + 1).  Ap [j] is
                  location in Ai or Ax of the first nonzero associated
                  with column j. Ap [ncol] is the total number of nonzeros
                  in the matrix

       (2) a packed array containing all the elements of A, both zeros and
           nonzeros.

       (3) triples format: 3 arrays containing
             rows - row     indices of nonzero entries
             cols - column  indices of nonzero entries
             vals - numerical values of nonzero entries

        For this example, it is easiest to use a packed array. Storing A by
        rows, we have */

    PASAFLOAT A [] = { 1, 2, 2, 0, 0, 1, 0, 0, 0, 1, 0,
                       0, 0, 0, 1, 2, 1, 1, 0, 0, 0, 0,
                       0, 0, 1, 0, 0, 0, 1, 1, 2, 1, 0,
                       1, 1, 1, 1, 1, 1, 1, 1, 1, 1,-1 } ;
    /* store the A pointer in the pasadata structure: */
    pasadata->A->by_rows = A ;

    /* Input the lower bounds for A*x */
    PASAFLOAT bl [] = { 2, 1, 1, 0 } ;
    pasadata->bl = bl ;
    /* pasadata->bu = bl since the upper and lower bounds are equal */
    pasadata->bu = bl ;

    /* the starting guess is the vector with all its entries 0.1 */
    PASAFLOAT *x = pasadata->x = (PASAFLOAT *) malloc (ncol*sizeof (PASAFLOAT));
    /* set x [i] = 0.1 for 0 <= i <= 9 */
    pasa_initx (x, 0.1, 10) ;
    /* set x [10] = 1 = x [0] + x [1] + ... + x [9] */
    x [10] = 1 ;

    /* lower bounds is the vector with all its entries 1.e-6 */
    PASAFLOAT *lo = pasadata->lo = (PASAFLOAT *) malloc (ncol*sizeof (PPFLOAT));
    pasa_initx (lo, 1.e-6, ncol) ;

    /* nothing is stored in pasadata->hi since there are no upper bounds */

    /* The routines to evaluate the function and the gradient appear below,
       while their prototypes are above.  Since it is often easy to
       evaluate the gradient at the same time as the function value,
       we also code a routine valgrad that simultaneously evaluates
       the function and its gradient.  The value and gradient routines
       are required, while the valgrad is optional. In some applications,
       the solution time is reduced when valgrad is provided. */
    pasadata->value = value ;
    pasadata->grad = grad ;
    pasadata->valgrad = valgrad ;

    /* The Hessian is also provided below, and it could potentially be
       be utilized by pasa if it is inserted in the pasadata structure:

    pasadata = hess ; */

    /* For this problem, the Hessian is not utilized by pass, even
       if it is provided. */

    /* Now solve the problem. */
    pasa (pasadata) ; 

    /* By default, pasadata->Parm->print_status = TRUE, so the status of
       the run will be printed. If this parameter was FALSE, then the user
       could manually print the status using pasa_print_status(pasadata);
       run status is also an output of pasa: int status = pasa (pasadata) ;
       The run was successful when status = 0.  If the run was successful,
       then the solution is stored in pasadata->x */
    pasa_printx (pasadata->x, ncol, "solution") ;

    /* The pasa_print_stats routine gives the statistics for all the
       routines used in the run, while pasa_print_parms outputs the
       parameter values of all the routines that were used during the
       solution process. Below, specific statistics are output: */
        PASAstat *pasastat = pasadata->Stats->pasa ;
        PPstat  *pprojstat = pasadata->Stats->pproj ;
        CGstat     *cgstat = pasadata->Stats->cg ;

        printf("\n ***************** Statistics for PASA Run **********"
               "**********\n\n") ;
        printf("Number of variables                     = %-10ld\n",
              (LONG) pasadata->ncol) ;
        printf("Number of linear constraints            = %-10ld\n\n",
              (LONG) pasadata->nrow) ;
        printf("Iterations of gradient projection (GP)  = %-10ld\n",
              (LONG) pasastat->gpit) ;
        printf("Iterations of active set GP             = %-10ld\n",
              (LONG) pasastat->agpit) ;
        printf("Iterations of conjugate gradient  (CG)  = %-10ld\n\n",
              (LONG) cgstat->iter) ;
        printf("Function evaluations in GP              = %-10ld\n",
              (LONG) pasastat->gpnf) ;
        printf("Function evaluations in active set GP   = %-10ld\n",
              (LONG) pasastat->agpnf) ;
        printf("Function evaluations in CG              = %-10ld\n\n",
              (LONG) cgstat->nfunc) ;
        printf("Gradient evaluations in GP              = %-10ld\n",
              (LONG) pasastat->gpng) ;
        printf("Gradient evaluations in active set GP   = %-10ld\n",
              (LONG) pasastat->agpng) ;
        printf("Gradient evaluations in CG              = %-10ld\n\n",
              (LONG) cgstat->ngrad) ;
        printf("Number of projections onto feasible set = %-10ld\n",
              (LONG) pasastat->nproject) ;
        printf("Number of Cholesky factorizations       = %-10i\n",
               pprojstat->nchols) ;
        printf("Sup-norm of projected gradient          = %-16.7e\n",
               pasastat->err) ;
        printf("Final objective value                   = %-16.7e\n",
               pasastat->f) ;
        printf("\n ******************************************************"
                   "************\n\n") ;
    /* Note: the default error tolerance is 1.e-6. It can be changed by
             choosing a different value for pasadata->Parms->pasa->grad_tol */

    /* free memory malloc'd above */
    free (lo) ;
    free (x) ;
    if ( rows != NULL ) free (rows) ;
    if ( cols != NULL ) free (cols) ;
    if ( vals != NULL ) free (vals) ;

    /* use pasa_terminate to free the pasadata structure and any memory
       that was malloc'd by pasa. */
    pasa_terminate (&pasadata) ;

    /* exit the program */
    return (0) ;
}

/* NOTE: By C convention, the first element of x is x [0] which corresponds
         to x_1 in the problem formulation, and x [10] corresponds to x_11 */
void value
(
    PASAFLOAT *val,
    PASAFLOAT   *x,
    PASAINT      n
)
{
    PASAFLOAT f ;
    PASAINT i ;

    /* Since the pasa iterates are feasible, x [10] = x [0] + ... + x [9]
       Thus the x_{11} term in f reduces to: */
    f = -x [10]*log (x [10]) ;
    /* now include the other terms in f: */
    for (i = 0; i < 10; i++)
    {
        f += x [i]*(c [i] + log (x [i])) ;
    }
    *val = f;
}

void grad
(
    PASAFLOAT *g,
    PASAFLOAT *x,
    PASAINT    n
)
{
    PASAFLOAT t ;
    PASAINT i ;

    t = 1 - log (x [10]) ;
    for (i = 0; i < 10; i++)
    {
        g [i] = c [i] + log (x [i]) + t ;
    }
    g [10] = -1 ;
}

void valgrad
(
    PASAFLOAT *val,
    PASAFLOAT   *g,
    PASAFLOAT   *x,
    PASAINT      n
)
{
    PASAFLOAT f, s, t ;
    PASAINT i ;

    t = 1 - log (x [10]) ;
    f = -x [10]*log (x [10]) ;
    for (i = 0; i < 10; i++)
    {
        PASAFLOAT const xi = x [i] ;
        s = c [i] + log (xi) ;
        f += xi*s ;
        g [i] = s + t ;
    }
    g [10] = -1 ;
    *val = f ;
}

/* The nonzero of the Hessian are on the main diagonal (1/x_i) and the
   last row and column (-1/x [10]). These nonzero are input using triples.
   The first row and column for triples are 1. Since the Hessian is symmetric,
   only the elements on the main diagonal and above are provided. */
void hess
(
    SOPT_matrix *H,
    SOPTFLOAT   *x,
    SOPTINT      n
)
{
   PASAFLOAT t ;
   CGINT i, j ;
    /* At the first call to hess, malloc arrays and save the row and column
       indices of the nonzeros. This is only done one time. In subsequent call
       to hess, we only provide the elements of the Hessian that could change
       when x changes.  The number of nonzeros elements to be provided
       is 21 corresponding to the 11 main diagonal elements (1/x_i) and the 10
       elements -(1/x_{11}) in the last column excluding the diagonal. Recall
       that x_{11} corresponds to x [10] in C. */
    t = 1/x [10] ;
    if ( H->nnz == EMPTY )
    {
        CGINT nnz ;
        nnz = 21 ;
        H->rows = rows = malloc (nnz*sizeof (CGINT)) ;
        H->cols = cols = malloc (nnz*sizeof (CGINT)) ;
        H->vals = vals = malloc (nnz*sizeof (CGFLOAT)) ;
        j = 0 ;
        /* NOTE: we use C-indexing for rows and columns; hence, the main
           diagonal and the last column of the Hessian are (i, i) and
           (i, 10), 0 <= i <= 10. */
        H->fortran = FALSE ; /* C indexing used */
        for (i = 0; i < 10; i++)
        {
            /* store the diagonal element */
            rows [j] = i ;
            cols [j] = i ;
            vals [j] = 1/x [i] ;
            j++ ;
            /* store the element in the last column */
            rows [j] = i ;
            cols [j] = 10 ; /* last column is 10 in C */
            vals [j] = -t ;  /* -1/x [10] */
            j++ ;
        }
        /* store the final element on the diagonal */
        rows [j] = 10 ;
        cols [j] = 10 ;
        vals [j] =  t ;
        /* the number of nonzeros on diagonal and above */
        H->nnz = nnz ;
        /* NOTE: H->sym set to TRUE in pasa_setup */
    }
    else /* update the elements of the Hessian that change when x changes */
    {
        vals = H->vals ;
        j = 0 ;
        for (i = 0; i < 10; i++)
        {
            vals [j] = 1/x [i] ;
            j++ ;
            vals [j] = -t ;
            j++ ;
        }
        vals [j] = t ;
    }
}
