/* ---------- pasa mex function ---------- */
/* This file contains the mex function for interfacing pasa with MATLAB */

/* ---------- Include header files ---------- */
#include "pasa_matlab.h"

/* ---------- Declare global variables ---------- */
mxArray *suiteopt_value ;
mxArray *suiteopt_grad ;
mxArray *suiteopt_valgrad ;
mxArray *suiteopt_hprod ;
mxArray *suiteopt_cg_hprod ;
mxArray *suiteopt_hess ;
int matlab_mallocd_hessian = FALSE ;

/* mex function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{
    /* Initialize input variables */
    int destroy_lambda = FALSE ;
    int create_lambda = FALSE ;
    double  *newx, maxits ;
    bool Aexists ;
    LONG i, ncol ;
    char *mystr ;
    mxArray *lhs [1], *rhs, *field ;
    const mxArray *options ;

    #if DEBUG_PASA_MEX
    mexPrintf ("Initialized variables\n") ;
    #endif

    /* If the user provided no input then print some options */
    if (nrhs == 0)
    {
        printf("\n\n  PASA (Polyhedral Active Set Algorithm) is designed\n"
               "    to solve problems of the form\n\n              "
               "min f(x)  subject to  bl <= Ax <= bu, lo <= x <= hi\n"); 
        /* Print additional information note for user */
        pasa_matlab_print_info () ;

        /* Nothing to solve, terminate program */
        return ;
    }

    /* Verify correct number of input arguments provided for PASA */
    if(nrhs > 1)
    {
        mexErrMsgIdAndTxt("MyToolbox:pasa:nrhs",
                          "pasa requires exactly one struct as an input.") ;
    }

    int string_input = mxIsChar(prhs[0]) ;

    /* Any string must be input as row vector */
    if ( string_input )
    {
        if ( mxGetM(prhs[0])!=1 )
        {
            mexErrMsgIdAndTxt("MATLAB:pasa:inputNotVector",
                    "pasa help keyword must be a row vector.") ;
        }

        /* Copy string from prhs[0] to string in C */
        mystr = mxArrayToString(prhs[0]);
        
        if( mystr == NULL ) 
        {
            mexErrMsgIdAndTxt("MATLAB:pasa:conversionFailed",
                  "Input could not be converted to string.");
        }
    }
    else /* not string input, must be structure */
    {
        /* Check if user provided nonempty struct as input */
        if ( !mxIsClass(prhs[0], "struct") || mxIsEmpty(prhs[0]) )
        {
            mexErrMsgIdAndTxt("MATLAB:pasa:conversionFailed",
                  "Input must be nonempty structure.");
        }
    }

    /* Verify correct number of output arguments provided for PASA */
    if( (nlhs < 1 || nlhs > 3) && !string_input )
    {  
        mexErrMsgIdAndTxt("MyToolbox:pasa:nlhs",
                          "pasa requires at least 1 output "
                          "and allows at most 3 outputs.") ;
    }

    #if DEBUG_PASA_MEX
    mexPrintf ("Finished prhs error messages\n") ;
    #endif

    /*------- Initialize pasa data struct -------- */
    PASAdata *pasadata = pasa_setup () ;

    /* If user provided no inputs then print default parameter values */
    if ( string_input )
    {
        /* check string against valid keywords */
        if (!strcmp(mystr, "readme"))
        {
            /* Print readme information about pasa */
            /* Copy prhs to non const mxArray */
            rhs = (mxArray*) prhs[0] ;
    
            /* Call MATLAB function help to print readme options */
            mexCallMATLAB(0, lhs, 1, &rhs, "help") ;
        }
        else if ((!strcmp(mystr, "all")) || (!strcmp(mystr, "allparms")))
        {
            /* Print default pasa parameter values and descriptions */
            pasa_print_parm (pasadata) ;
    
            /* Print default pproj parameter values and descriptions */
            pproj_print_parm (pasadata->ppdata) ;

            /* Print default cg parameter values and descriptions */
            cg_print_parm (pasadata->cgdata) ;

            /* Print default napheap parameter values and descriptions */
            napheap_print_parm (pasadata->napdata) ;
        }
        else if (!strcmp(mystr, "parm"))
        {
            /* Print default pasa parameter values and descriptions */
            pasa_print_parm (pasadata) ;
        }
        else if (!strcmp(mystr, "pproj"))
        {
            /* Print default pproj parameter values and descriptions */
            pproj_print_parm (pasadata->ppdata) ;
        }
        else if (!strcmp(mystr, "cg"))
        {
            /* Print default pproj parameter values and descriptions */
            cg_print_parm (pasadata->cgdata) ;
        }
        else if (!strcmp(mystr, "napheap"))
        {
            /* Print default pproj parameter values and descriptions */
            napheap_print_parm (pasadata->napdata) ;
        }
        else if (!strcmp(mystr, "demo"))
        {
            /* Open demo file for pasa */
            /* Copy prhs to non const mxArray */
            rhs = (mxArray*) prhs[0] ;
    
            /* Call MATLAB function help to print readme options */
            mexCallMATLAB(0, lhs, 1, &rhs, "open") ;
        }
        else if (!strcmp(mystr, "demoQP"))
        {
            /* Open demo file for pasa */
            /* Copy prhs to non const mxArray */
            rhs = (mxArray*) prhs[0] ;
    
            /* Call MATLAB function help to view file */
            mexCallMATLAB(0, lhs, 1, &rhs, "open") ;
        }
        else if (!strcmp(mystr, "demoOC"))
        {
            /* Open demo file for pasa */
            /* Copy prhs to non const mxArray */
            rhs = (mxArray*) prhs[0] ;
    
            /* Call MATLAB function help to view file */
            mexCallMATLAB(0, lhs, 1, &rhs, "open") ;
        }
        else /* Invalid option string provided for pasa */ 
        {
            printf("\n Invalid keyword entered for pasa information.\n") ;
        }

        /* Print additional information note for user */
        pasa_matlab_print_info () ;
        
        pasa_terminate (&pasadata) ;

        /* Exit program as there is nothing to evaluate */
        return ;
    }

    #if DEBUG_PASA_MEX
    mexPrintf("Loading problem data struct\n") ;
    #endif

    /* Copy input to options */
    options = prhs[0] ;

    /* Initialize Aexists to FALSE (no linear inequality constraints) */
    Aexists = FALSE ; 

    /* --------------------- Import parameters ---------------------- */
    /* Import any custom parameters specified by users */

    #if DEBUG_PASA_MEX
    mexPrintf("Loading problem parameters from struct\n") ;
    #endif

    /* PASAparms: Import PASA, PPROJ, CG, and NAPHEAP parameters --- */
    pasa_matlab_get_parms (options, pasadata->Parms) ;

    /* ----------- Import problem options provided in input struct ------- */
    /* Detailed description of input options found in the setup function
       in PASA/Source/pasa.c */

    /* ncol: Problem dimension */
    suiteopt_matlab_get_problem_dimension (options, &pasadata->ncol) ;
    /* nrow, ((Ap, Ai, Ax) or (a)), and
       (bl, bu): Polyhedral/napsack constraint */
    suiteopt_matlab_get_polyhedral_constraints (options, pasadata) ;

    /* lo, hi: Bound constraints lo <= x <= hi */
    suiteopt_matlab_get_bound_constraints (options, pasadata) ;

    /* c: Linear term in quadratic cost function */
    pasadata->c = suiteopt_matlab_get_float (options, &(pasadata->ncol), "c") ;

    /* d: Diagonal of Hessian in napsack problem (if NAPSACK) --- */
    pasadata->d = suiteopt_matlab_get_float (options, &(pasadata->ncol), "d") ;

    /* y: Point to project onto polyhedron (if PROJ) */
    pasadata->y = suiteopt_matlab_get_float (options, &(pasadata->ncol), "y") ;

    /* --- x (Solution): Allocate memory if not provided --- */
    if ((field = mxGetField(options, 0, "x")) != NULL)
    {
        if (mxIsClass(field, "double"))
        {
            /* -------- Configure initial output values  -------- */
            /* Copy initial guess for x into first output plhs[0] */
            plhs[0] = mxDuplicateArray(field) ;

            #if DEBUG_CGDESCENT_MEX
            mexPrintf("Successfully set plhs[0].\n") ;
            #endif

            /* Copy value for x */
            pasadata->x = mxGetPr(plhs[0]) ;

            #if DEBUG_CGDESCENT_MEX
            mexPrintf("Successfully imported initial guess x.\n") ;
            #endif

            i = (LONG) PASAMAX(mxGetN (field), mxGetM (field)) ;
            /* Set problem dimension if n not provided */
            if ( pasadata->ncol <= 0 )
            {   
                pasadata->ncol = i ;
            }
            else if ( pasadata->ncol > i ) /* print error and exit */
            {
                mexPrintf("The dimension of the starting guess is %ld, "
                          "which is inconsistent\n"
                          "with the specified problem dimension given by "
                          "pasadata.ncol = %ld\n", i, pasadata->ncol) ;
                mexErrMsgTxt("Terminating pasa.\n") ;
            }
        }
        else /* x must be of type double, print error message */
        {
            mexPrintf("Initial guess lambda provided to pasa "
                      "must be of type double.\n") ;
            mexErrMsgTxt("Terminating pasa.\n") ;
        }
    }
    else if (pasadata->ncol > 0) /* x not provided but dimension provided */
    {
        #if DEBUG_CGDESCENT_MEX
        mexPrintf("Initial guess x not provided.\n");
        #endif

        /* Initialize pointer for the first output of plhs */
        plhs[0] = mxCreateDoubleMatrix (pasadata->ncol, 1, mxREAL) ;
        /* Copy value for x */
        pasadata->x = mxGetPr(plhs[0]) ;
        pasa_initx (pasadata->x, PASAZERO, pasadata->ncol) ;

        #if DEBUG_CGDESCENT_MEX
        mexPrintf("Allocated memory for x.\n");
        #endif
    }
    else /* x not provided and pasadata->ncol not provided, Print error */
    {
        mexPrintf("Initial guess x and problem dimension "
                  "not provided to pasa.\n") ;
        mexErrMsgTxt("Terminating pasa.\n") ;
    }

    if ( pasadata->ncol == EMPTY )
    {
        mexPrintf("PASA was unable to extract a problem dimension\n") ;
        mexErrMsgTxt("Terminating execution.\n") ;
    }

    /* if there are constraints, then check if guess for lambda is present */
    if (pasadata->nrow > 0)
    {
        /* lambda: Initial guess for multiplier --- */
        if ((field = mxGetField(options, 0, "lambda")) != NULL)
        {
            if (mxIsClass(field, "double"))
            {
                i = (LONG) PASAMAX(mxGetN (field), mxGetM (field)) ;
                if ( i != pasadata->nrow )
                {
                    mexPrintf ("Dimension %li of the multiplier starting guess "
                               "lambda is inconsistent with the number of\n"
                               "rows %li in the constraint polyhedron\n",
                               i, pasadata->nrow) ;
                    mexErrMsgTxt("Terminating pasa.\n") ;
                }
                if ( nlhs >= 3 ) /* use plhs[2] to stored the starting lambda */
                {
                    plhs [2] = mxDuplicateArray(field) ;

                    #if DEBUG_CGDESCENT_MEX
                    mexPrintf("Successfully set plhs[2].\n") ;
                    #endif

                    /* Copy value for lambda */
                    pasadata->lambda = mxGetPr (plhs [2]) ;

                    #if DEBUG_CGDESCENT_MEX
                    mexPrintf("Successfully imported initial guess lambda.\n") ;
                    #endif
                }
                else /* lambda is not output */
                {
                    /* Copy value for lambda */
                    pasadata->lambda = mxGetPr (mxDuplicateArray(field)) ;
                    /* lambda is used, but not returned */
                    destroy_lambda = TRUE ;
                }
            }
            else /* lambda must be of type double, print error message */
            {
                mexPrintf("Initial guess lambda provided to pasa "
                          "must be of type double.\n") ;
                mexErrMsgTxt("Terminating pasa.\n") ;
            }
        }
        else if ( nlhs >= 3 )
        {
            /* starting guess for lambda not provided, but final lambda output*/
            create_lambda = TRUE ;
        }
    }

    /* objective: Function handle for objective */
    suiteopt_value = suiteopt_matlab_get_func (options, 1, 1, "value") ;
    if (suiteopt_value != NULL) pasadata->value = suiteopt_matlab_value ;

    /* gradient: Function handle for gradient */
    suiteopt_grad = suiteopt_matlab_get_func (options, 1, 1, "grad") ;
    if (suiteopt_grad != NULL) pasadata->grad = suiteopt_matlab_grad ;

    /* valgrad: Function handle for objective & gradient */
    suiteopt_valgrad = suiteopt_matlab_get_func (options, 1, 2, "valgrad") ;
    if (suiteopt_valgrad != NULL) pasadata->valgrad = suiteopt_matlab_valgrad ;

    suiteopt_matlab_get_Hessian_matrix (options, pasadata) ;
    
    /* hprod: Function handle for PASA Hessian times vector */
    suiteopt_hprod = suiteopt_matlab_get_func (options, 2, 1, "hprod") ;
    if (suiteopt_hprod != NULL) pasadata->hprod = suiteopt_matlab_hprod ;

    /* cg_hprod: Function handle for CG Hessian times vector */
    suiteopt_cg_hprod = suiteopt_matlab_get_func (options, 1, 1, "cg_hprod") ;
    if (suiteopt_cg_hprod != NULL) pasadata->cg_hprod = suiteopt_matlab_cghprod;

    /* hessian: Function handle for the evaluation of the objective Hessian */
    suiteopt_hess = suiteopt_matlab_get_func (options, 1, 1, "hessian") ;
    if ( suiteopt_hess != NULL ) pasadata->hessian = suiteopt_matlab_hessian ;

    #if DEBUG_PASA_MEX
    mexPrintf("Completed importing all prhs.\n") ;
    #endif

    /* ------- Check if problem dimension was not provided ------- */
    if (pasadata->ncol <= 0)
    {
        mexPrintf ("pasa unable to determine problem dimension based "
                   "on provided input.\n") ;
        mexErrMsgTxt("Terminating pasa.\n") ;
    }

    #if DEBUG_PASA_MEX
    /* Print user provided data */
    pasa_matlab_print_data (pasadata) ;
    #endif

    /* -------- Initialize pproj statistics to default -------- */
    pproj_initstat (pasadata->Stats->pproj) ;

    /* -------- Run pasa on problem -------- */
    pasa (pasadata) ;

    #if DEBUG_PASA_MEX
    mexPrintf("Finished running pasa.\n") ;
    #endif

    /* -------- Store output values -------- */
    #if DEBUG_PASA_MEX
    mexPrintf("Storing pasa problem solution.\n") ;
    #endif

    plhs[0] = mxCreateDoubleMatrix(1, pasadata->ncol, mxREAL) ;
    newx = mxGetPr(plhs[0]) ;

    #if DEBUG_PASA_MEX
    mexPrintf("Copying x to output array.\n") ;
    #endif

    pasa_copyx (newx, pasadata->x, pasadata->ncol) ;

    #if DEBUG_PASA_MEX
    mexPrintf("x copied to output successfully.\n") ;
    #endif

    /* -------- Store any remaining output values -------- */
    if (nlhs >= 2)
    {
        #if DEBUG_PASA_MEX
        mexPrintf("Storing pasa statistics for output.\n") ;
        #endif

        #if DEBUG_PASA_MEX
        mexPrintf("upasastats->pproj->maxdepth = %i\n", 
                  pasadata->Stats->pproj->maxdepth) ;
        #endif

        /* Determine value for Aexists */
        Aexists = (pasadata->A != NULL) ;

        /* Store pasa statistics in second output */
        pasa_matlab_get_stats (&plhs[1], pasadata->Stats, Aexists) ;

        #if DEBUG_PASA_MEX
        mexPrintf("Stored pasa statistics successfully.\n") ;
        #endif
    }
    /* If specified, store final lambda for third output. */
    if ( destroy_lambda )
    {
        mxFree (pasadata->lambda) ;
    }
    else if ( create_lambda )
    {
        #if DEBUG_PASA_MEX
        mexPrintf("Initializing output for lambda.\n") ;
        #endif
   
        plhs[2] = mxCreateDoubleMatrix(1, pasadata->nrow, mxREAL) ;
    
        #if DEBUG_PASA_MEX
        mexPrintf("Copying lambda to output array.\n") ;
        #endif
  
        pasa_copyx (mxGetPr (plhs [2]), pasadata->lambda, pasadata->nrow) ;
 
        #if DEBUG_PASA_MEX
        mexPrintf("Lambda copied to output successfully.\n") ;
        #endif
    }

    /* -------- Free memory allocated in mex function -------- */
    #if DEBUG_PASA_MEX
    mexPrintf("Freeing memory allocated by pasa_setup.\n") ;
    #endif

    if ( matlab_mallocd_hessian )
    {
        if ( pasadata->H->rows != NULL ) mxFree (pasadata->H->rows) ;
        if ( pasadata->H->cols != NULL ) mxFree (pasadata->H->cols) ;
        if ( pasadata->H->vals != NULL ) mxFree (pasadata->H->vals) ;
    }

    pasa_terminate (&pasadata) ;

    #if DEBUG_PASA_MEX
    mexPrintf("Exiting pasa.\n") ;
    #endif

    return ;
}
