% ================================ PASA README =================================
% The PASA (Polyhedral Active Set Algorithm) is designed to solve problems of
% the form
%
%              min f(x)  subject to  bl <= A*x <= bu, lo <= x <= hi
%
% This file explains how to install and use pasa in MATLAB. The file is best
% viewed by starting MATLAB and typing the command
%
%    >> pasa readme
%
% ========================= Compiling PASA for MATLAB ==========================
%  To pasa mex function has not yet been compiled, then startup MATLAB in
%  this directory and enter the following command in MATLAB:
%
%    >> make
%
% ============================== PASA quick use ================================
%  The data describing the problem should be stored in a structure denoted
%  pasadata. The problem is solved using a statement of the form
%
%   >> [x, stats, lambda] = pasa (pasadata) ;
%
%  The only required output is x, the problem solution; both the statistics
%  structure (stats), and the multiplier (lambda) associated with the
%  constraints bl <= A*x <= bu are optional. If output lambda is desired, then
%  stats must be included.
%
% ================================ PASA inputs =================================
%  The elements of the pasadata structure are the following:
%
%    pasadata.A         : constraint matrix (either sparse or dense format)
%    pasadata.bl        : lower bounds for A*x
%    pasadata.bu        : upper bounds for A*x
%    pasadata.H         : Hessian of objective if QP (sparse or dense format)
%    pasadata.c         : Linear term in objective if LP or QP (dense vector)
%    pasadata.ncol      : dimension of x
%    pasadata.lo        : lower bounds for x
%    pasadata.hi        : upper bounds for x
%    pasadata.x         : guess for the solution x
%    pasadata.lambda    : guess for multiplier of constraint bl <= A*x <= bu
%    pasadata.value     : Function handle for evaluating objective function
%    pasadata.grad      : Function handle for gradient of the objective function
%    pasadata.valgrad   : Function handle for objective and its gradient
%    pasadata.hessian   : Function handle for the Hessian of the objective
%    pasadata.hprod     : Function handle for evaluating hessian times vector
%    pasadata.a         : When A has one row, input as dense row pasadata.a
%    pasadata.d         : f(x) = .5x'*D*x + c'*x, D diagonal matrix, d = diag(D)
%    pasadata.y         : f(x) = ||y-x||,
%    pasadata.cg_hprod  : Function handle for hessian times vector in
%                         unconstrained optimization
%
%  If any elements of the structure do not exist or are unavailable
%  then they should not be specified. For example, if a starting
%  guess for the solution is not known, then omit pasadata.x and
%  pasa will use x = 0 for the starting guess.
%
%  Finally, any of the parameters used by pasa, pproj, cg_descent, or napheap
%  can appear in the pasadata structure. They are specified using a statement
%  of the form
%
%  >> pasadata.SS.PP = VV ;
%
%  where SS is any of the solvers pasa, pproj, napheap, or cg, PP is the
%  name of the parameter, and VV is the value assigned to the parameter.
%  A list of the default parameter values is obtained by commands of the
%  following form:
%
%      'pasa all'     - List of all default parameters values
%      'pasa parm'    - List of default pasa parameter values
%      'pasa pproj'   - List of default pproj parameter values
%      'pasa cg'      - List of default cg_descent parameter values
%      'pasa napheap' - List of default napheap parameter values
%
%  For example, the stopping criterion in pasa is set to 1.e-8 by
%  >> pasadata.pasa.grad_tol = 1.e-8 ;
%
%  Nearly all the elements of the pasadata structure are optional; the code
%  will attempt to deduce the value of missing elements. If ncol is not given,
%  then the deduced value for ncol is the minimum of the column dimension of A
%  and the lengths of c, lo, hi, x, and y, if they exist. If the user wishes
%  to enforce the constraint bl <= A*x <= bu, then A and bl or bu are all
%  required, A must be provided in sparse format, and the lengths of bl and bu
%  must match the number of rows in A.  If pasadata.lo or pasadata.hi are
%  not given, then it is assumed that the respective constraints lo <= x
%  or x <= hi do not exist.  pasadata->bl and pasadata->bu are treated
%  similarly. If pasadata.x or pasadata.lambda are not given, then they
%  are taken to be zero.
%
% ============================= Non-Quadratic Mode =============================
%  Here we outline which function handles are required and optional when
%  using the non-quadratic mode of PASA.
%
%   ---------------------- Required Function Handles --------------------------
%    pasadata.value   : Function handle for evaluating objective function
%                       f = val(x)
%    pasadata.grad    : Function handle for gradient of the objective function
%                       g = grad(x)
%   ---------------------- Optional Function Handle ---------------------------
%    pasadata.valgrad : Function handle for objective and gradient
%                       [f, g] = valgrad(x)
%    pasadata.hessian : Function handle for evaluating objective Hessian
%                       H = hess(x)
%                       H = [rows, cols, vals] in triples format. Only provide
%                           elements on main diagonal and above. vals(i) is the
%                           nonzero at location (rows(i),cols(i)) in the Hessian
%
%  Note that value, grad, and hessian are functions with one input and one 
%  output and valgrad is a function with one input and two outputs. 
%
% ====================== Quadratic or Linear Mode ==============================
%  If the objective is either quadratic or linear, then the user can provide the
%  linear part of objective in pasadata.c as a dense vector. For a quadratic,
%  the Hessian can be provided in pasadata.H as either a sparse or a dense
%  matrix. Alternatively, the user can provide a routine that evaluates the
%  product between the Hessian and a vector. If the optimization problem is
%  unconstrained, then the function handle pasadata.cg_prod can be employed,
%  in which case cgprod(x) outputs the product p = H*x between the Hessian
%  H and the vector x. If the optimization problem has constraints, then
%  pasadata.hprod should be utilized, in which case hprod (d, ifree) output
%  the product p = H(:, ifree)*d where ifree is a vector containing column
%  numbers of the Hessian, d is a vector of the same length as ifree, and
%  p = sum_i H(:, ifree_i)*d_i  (see example in demoQP).
